// Rectifier.cpp
//

// LDraw code "rectifier"
// - condenses triangles that share a common side and are in the same plane into a quad.
// - Finds completely flat parallelogram quads that are bordered by edge lines and replace them with rect primitives

// Usage: Rectifier [-q] [-r] [-c] [-t <value>] LdrawFileIn LdrawFileOut

// -q: Don't condense triangles into quads
// -r: don't substitute rect primitives
// -c: color new quads in yellow and new rects in blue
// -t <value>: coplanarity threshold for triangle to quad merging (default 0.95 degrees)

// Philippe "Philo" Hurbain - 2008 - www.philohome.com

// Contains softSurfer Point - Line distance calculation (adaptation)
// Copyright 2001, softSurfer (www.softsurfer.com)
// This code may be freely used and modified for any purpose
// providing that this copyright notice is included with it.
// SoftSurfer makes no warranty for this code, and cannot be held
// liable for any real or imagined damage resulting from its use.
// Users of this code must verify correctness for their application.

#include "stdafx.h"

#include <math.h>
#include <string.h>
#include <stdlib.h>

#define USE_EPSILON_TEST TRUE  
#define EPSILON 0.00001

// Flat angle threshold: 0.95 degrees
double AngleThr = 0.95;


/* Null vector */

double null[3] = {0,0,0};

/* some macros */

#define CROSS(dest,v1,v2)                      \
              dest[0]=v1[1]*v2[2]-v1[2]*v2[1]; \
              dest[1]=v1[2]*v2[0]-v1[0]*v2[2]; \
              dest[2]=v1[0]*v2[1]-v1[1]*v2[0];

#define DOT(v1,v2) (v1[0]*v2[0]+v1[1]*v2[1]+v1[2]*v2[2])

#define SUB(dest,v1,v2) dest[0]=v1[0]-v2[0]; dest[1]=v1[1]-v2[1]; dest[2]=v1[2]-v2[2]; 

#define ADD(dest,v1,v2) dest[0]=v1[0]+v2[0]; dest[1]=v1[1]+v2[1]; dest[2]=v1[2]+v2[2]; 

#define MULT(dest,v,factor) dest[0]=factor*v[0]; dest[1]=factor*v[1]; dest[2]=factor*v[2];

#define SET(dest,src) dest[0]=src[0]; dest[1]=src[1]; dest[2]=src[2];

#define MANHATTAN(v1, v2) (fabs(v1[0]-v2[0]) + fabs(v1[1]-v2[1]) + fabs(v1[2]-v2[2]))

#define NORM(v1) (sqrt(DOT(v1,v1)))

#define DIST(p1, p2) (sqrt((p1[0]-p2[0])*(p1[0]-p2[0])+(p1[1]-p2[1])*(p1[1]-p2[1])+(p1[2]-p2[2])*(p1[2]-p2[2])))


#define ROTTRI(tri) { double temp[3]; SET(temp,tri.a); SET(tri.a,tri.b); SET(tri.b,tri.c); SET(tri.c,temp); }

#define ROTQUAD(quad) { double temp[3]; SET(temp,quad.a); SET(quad.a,quad.b); SET(quad.b,quad.c); SET(quad.c,quad.d); SET(quad.d,temp); }

#define SWAP(p1,p2) { double temp[3]; SET(temp,p1); SET(p1,p2); SET(p2,temp); }


// Check if two coplanar triangles have the same winding
int SameWinding (double T1a[3], double T1b[3], double T1c[3], double T2a[3], double T2b[3], double T2c[3])
{
	double u1[3], v1[3], u2[3], v2[3];
	double n1[3], n2[3];
	SUB(u1, T1b, T1a);
	SUB(v1, T1c, T1a);
	SUB(u2, T2b, T2a);
	SUB(v2, T2c, T2a);
	CROSS(n1, u1, v1);
	CROSS(n2, u2, v2);
	if(DOT(n1,n2)>0) return true; 
	else return false;
}


// Calculate "flatness" of angle p1p2p3 (_NOT_ normalized!)
double flatness (double p1[3], double p2[3], double p3[3])
{
	double v[3], w[3], x[3];
	SUB(v,p2,p1);
	SUB(w,p2,p3);
	CROSS(x,v,w);
	return (DOT(x,x));
}

	
// dist_Point_to_Line(): get the distance of a point to a line.
//    Input:  a Point P and a Line L (in any dimension)
//    Return: the shortest distance from P to L

double dist_Point_to_Line( double P[3], double L1[3], double L2[3])
{
	double v[3], w[3], Pb[3];
	SUB (v, L2, L1);
	SUB (w, P, L1);

    double c1 = DOT(w,v);
    double c2 = DOT(v,v);
    double b = c1 / c2;

	MULT (w,v,b);
	ADD (Pb, L1, w);
    return DIST(P, Pb);
}

// Calculate angle of 3 points, U1 is the summit.
double Angle(double U0[3],double U1[3],double U2[3])
{
	double U01[3], U12[3];
	double temp[3], result;
	SUB(U01, U1, U0);
	SUB(U12, U2, U1);
	if((DOT(U01,U01)==0) || (DOT(U12,U12)==0)) return 0.;
	CROSS(temp, U01, U12);
	MULT(temp, temp, 1/(NORM(U01))/(NORM(U12)));
	if(NORM(temp)>(1-EPSILON)) result = 90;
	else result=180/3.1415926*asin((NORM(temp)));
	if(DOT(U01, U12) > 0 ) result = 180-result;
	return(result);

}

// Check if a quad is convex and doesn't have aligned vertices.
int Convex(double U0[3],double U1[3],double U2[3],double U3[3])
{
	double U01[3], U12[3], U23[3], U30[3];
	double T0[3], T1[3], T2[3], T3[3];
	int cnt=0;
	SUB(U01, U1, U0);
	SUB(U12, U2, U1);
	SUB(U23, U3, U2);
	SUB(U30, U0, U3);

	CROSS(T0, U30, U01);
	CROSS(T1, U01, U12);
	CROSS(T2, U12, U23);
	CROSS(T3, U23, U30);

	// Colinear Vertices
	if((Angle(U0, U1, U2)>179.9) || (Angle(U0, U1, U2)<0.025)) return false;
	if((Angle(U1, U2, U3)>179.9) || (Angle(U1, U2, U3)<0.025)) return false;
	if((Angle(U2, U3, U0)>179.9) || (Angle(U2, U3, U0)<0.025)) return false;
	if((Angle(U3, U0, U1)>179.9) || (Angle(U3, U0, U1)<0.025)) return false;

	// Check if convex
	if(DOT(T0, T1) > 0) cnt++; else cnt--;
	if(DOT(T1, T2) > 0) cnt++; else cnt--;
	if(DOT(T2, T3) > 0) cnt++; else cnt--;
	if(DOT(T3, T0) > 0) cnt++; else cnt--;

	if(abs(cnt) == 4) return true;
	else return false;
}

// Check if a quad is planar enough
int Planar( double U0[3],double U1[3],double U2[3],double U3[3])
{

	double a[3], b[3], c[3], d[4];
	double temp[3], n1[3], n2[3];
	
	SUB(a,U0, U1);
	SUB(b,U1, U2);
	SUB(c,U2, U3);
	SUB(d,U3, U0);
	CROSS(temp, a ,d);
	MULT(n1,temp,1/NORM(temp));
	CROSS(temp, b ,c);
	MULT(n2,temp,1/NORM(temp));
	CROSS(temp, n1, n2);
	if (180/3.1415926*asin((NORM(temp))) > AngleThr) return(false);

	CROSS(temp, a ,b);
	MULT(n1,temp,1/NORM(temp));
	CROSS(temp, d ,c);
	MULT(n2,temp,1/NORM(temp));
	CROSS(temp, n1, n2);
	if(180/3.1415926*asin((NORM(temp))) > AngleThr) return(false);
	return true;
}

// Greatest common divisor of 3 values
int gcd (int a, int b, int c )
{
	int d;
	while ( a != 0 ) 
	{
    d = a; a = b%a;  b = d;
	}
	while ( b != 0 ) 
	{
    d = b; b = c%b;  c = d;
	}
	return abs(c);
}


void usage(void)
{

	printf ("Usage: Rectifier [-q] [-r] LdrawFileIn LdrawFileOut\n");
	printf ("Options\n");
	printf ("  -q: don't condense triangles into quads\n");
	printf ("  -r: don't substitute rect primitives\n");
	printf ("  -a: no replacement of quads that have adjacent condlines\n");
	printf ("  -c: color new quads in yellow and new rects in blue\n");
	printf ("  -t <value>: coplanarity threshold for triangle to quad merging (default 0.95 degrees)\n");
	printf("Press <Enter> to quit");
	getchar();
}

#define MAX_IN 5000
#define SMALL 0.01



typedef struct
{
	int type;
	int color;
	int hex_color;
	double a[3];
	double b[3];
	double c[3];
	double d[3];
	int rect;
	int bfc;
} LDrawLine;

#define LDRAWCPY(dest,src)	\
{							\
	SET(dest.a, src.a);		\
	SET(dest.b, src.b);		\
	SET(dest.c, src.c);		\
	SET(dest.d, src.d);		\
	dest.type = src.type;	\
	dest.color = src.color;	\
	dest.rect = src.rect;	\
	dest.bfc = src.bfc;		\
}

// LDrawLine In[MAX_IN];

// typedef LDrawLine *LDLptr;

// LDLptr test;

	LDrawLine *In;

int main(int argc, char* argv[])
{

	FILE *LdrawIn;
	FILE *LdrawOut;

	char buf[1024];
	int type;

	int NumIn;
	int NumInMax;
	int SubstLine=0, SubstQuad=0;
	int SubstTri=0, SubstCond=0;

	int i,j;

	int ccw=true;
	int noquad=false;
	int norect=false;
	int color=false;
	bool noadjacent=false;

	printf("\nLdraw Rectifier v1.9 - by Philo");
	printf("\n---------------------------------\n");

	if(argc < 3) 		
	{
		usage();
		return 1;
	}

	for(i=1; i<argc-2; i++)
	{
		if(argv[i][0] != '-') 
		{
			usage();
			return 1;
		}
		switch (argv[i][1])
		{
		case 'q':
			noquad = true;
			printf("No replacement of triangles with quads\n");
			break;
		case 'r':
			norect = true;
			printf("No replacement of quads with rect primitives\n");
			break;
		case 'a':
			noadjacent=true;
			printf("No replacement of quads that have adjacent condlines\n");
			break;
		case 'c':
			color = true;
			printf("Coloring of substituted elements\n");
			break;
		case 't':
			i++;
			sscanf(argv[i], "%lf",&AngleThr);
			printf("Flat angle threshold: %.4lg", AngleThr);
			break;
		default:
			usage();
			return 1;
		}
	}


	LdrawIn = fopen (argv[argc-2], "rt");
	if(LdrawIn == NULL)
	{
		printf("Can't open file\n%s\n", argv[argc-3]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	LdrawOut = fopen (argv[argc-1], "wt");
	if(LdrawOut == NULL)
	{
		printf("Can't create file\n%s\n", argv[argc-1]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	printf("\nReading Input File...\n");
	NumIn=0;
	// Get number of triangles of infile1.
	while (fgets(buf, 1024, LdrawIn)) NumIn++;

	printf("  %d lines in input file\n", NumIn);
	rewind(LdrawIn);
	In=(LDrawLine*)malloc(NumIn*sizeof(LDrawLine));
	if(In==NULL)
	{
		printf("Can't allocate memory\n");
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}
	NumIn=0;

	while (fgets(buf, 1024, LdrawIn))
	{
		type = -1;
		sscanf(buf, "%d", &type);
		
		switch(type)
		{
		case -1:
		case 0:
			if(((strstr(buf,"BFC")!=NULL)||(strstr(buf,"bfc")!=NULL))&&((strstr(buf,"CERTIFY")!=NULL)||(strstr(buf,"certify")!=NULL)))
			{
				if((strstr(buf,"CCW")!=NULL)||(strstr(buf,"ccw")!=NULL))
				{
					ccw=true;
				}
				else if((strstr(buf,"CW")!=NULL)||(strstr(buf,"cw")!=NULL))
				{
					ccw=false;
				}
			}

		case 1:
			In[NumIn].type = 0;
			NumIn++;
			break;
		case 2:
			if (strstr(buf,"0x2")!=NULL)
			{
				sscanf (buf, "%d %x %lf %lf %lf %lf %lf %lf", &In[NumIn].type, &In[NumIn].hex_color, 
					&In[NumIn].a[0], &In[NumIn].a[1], &In[NumIn].a[2], 
					&In[NumIn].b[0], &In[NumIn].b[1], &In[NumIn].b[2]);
				In[NumIn].color = -1;
			}
			else
			{
				sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf", &In[NumIn].type, &In[NumIn].color, 
					&In[NumIn].a[0], &In[NumIn].a[1], &In[NumIn].a[2], 
					&In[NumIn].b[0], &In[NumIn].b[1], &In[NumIn].b[2]);
			}
			In[NumIn].bfc=ccw;
			NumIn++;
			break;
		case 3:
			if (strstr(buf,"0x2")!=NULL)
			{
				sscanf (buf, "%d %x %lf %lf %lf %lf %lf %lf %lf %lf %lf", &In[NumIn].type, &In[NumIn].hex_color, 
					&In[NumIn].a[0], &In[NumIn].a[1], &In[NumIn].a[2], 
					&In[NumIn].b[0], &In[NumIn].b[1], &In[NumIn].b[2],
					&In[NumIn].c[0], &In[NumIn].c[1], &In[NumIn].c[2]);
				In[NumIn].color = -1;
			}
			else
			{
				sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf", &In[NumIn].type, &In[NumIn].color, 
					&In[NumIn].a[0], &In[NumIn].a[1], &In[NumIn].a[2], 
					&In[NumIn].b[0], &In[NumIn].b[1], &In[NumIn].b[2],
					&In[NumIn].c[0], &In[NumIn].c[1], &In[NumIn].c[2]);
			}
			In[NumIn].bfc=ccw;
			NumIn++;
			break;
		case 4:
		case 5:
			if (strstr(buf,"0x2")!=NULL)
			{
				sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf", &In[NumIn].type, &In[NumIn].hex_color, 
					&In[NumIn].a[0], &In[NumIn].a[1], &In[NumIn].a[2], 
					&In[NumIn].b[0], &In[NumIn].b[1], &In[NumIn].b[2],
					&In[NumIn].c[0], &In[NumIn].c[1], &In[NumIn].c[2],
					&In[NumIn].d[0], &In[NumIn].d[1], &In[NumIn].d[2]);
				In[NumIn].color = -1;
			}
			else
			{
				sscanf (buf, "%d %d %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf %lf", &In[NumIn].type, &In[NumIn].color, 
					&In[NumIn].a[0], &In[NumIn].a[1], &In[NumIn].a[2], 
					&In[NumIn].b[0], &In[NumIn].b[1], &In[NumIn].b[2],
					&In[NumIn].c[0], &In[NumIn].c[1], &In[NumIn].c[2],
					&In[NumIn].d[0], &In[NumIn].d[1], &In[NumIn].d[2]);
			}	
			In[NumIn].bfc=ccw;
			NumIn++;
			break;
		default :
			break;
		}
	}
	NumInMax=NumIn;	

	fclose(LdrawIn);

	if(noquad==false)
	{
		printf("Substituting quads...\n");
		for(i=0; i<NumInMax; i++)
		{
			int j, k, l, m;


			if(In[i].type == 3)
			{
				int found=false;
				for (j=i+1; j<NumInMax; j++)
				{
					if(In[j].type != 3) continue;
					if(In[i].color != In[j].color) continue;
					if (In[i].color == -1)
					{
						if(In[i].hex_color != In[j].hex_color) continue;
					}
					for(k=0; k<3; k++)
					{
						for(l=0; l<3; l++)
						{
							if((DIST(In[i].a, In[j].a) < SMALL) && (DIST(In[i].c, In[j].b) < SMALL))
							{
								SET(In[i].d, In[j].c);
								if((Convex(In[i].a, In[i].b, In[i].c, In[i].d)) && (Planar(In[i].a, In[i].b, In[i].c, In[i].d)))
								{
									In[i].type=4;
									In[j].type=-1;
									if(color) In[i].color = 14;
									for(m=0; m<NumInMax; m++)
									{
										if(In[m].type == 5)
										{
											if(((DIST(In[i].a, In[m].a) < SMALL) && (DIST(In[i].c, In[m].b) < SMALL)) || ((DIST(In[i].a, In[m].b) < SMALL) && (DIST(In[i].c, In[m].a) < SMALL)))
											{
												In[m].type = -1;
												SubstCond++;
												break;
											}
										}
									}
									found=true;
									SubstTri++;
									break;
								}
							}
							ROTTRI(In[j]);
						}
						if(found) break;
						ROTTRI(In[i]);
					}
					if(found) break;
				}
			}
		}

	if(SubstTri > 0) printf("%d triangles(s) and %d conditional line(s) were replaced with %d quad(s)\n", SubstTri*2, SubstCond, SubstTri);
	}


	if(norect==false)
	{
		printf("Substituting rect primitives...\n");
		for(i=0; i<NumInMax; i++)
		{
			if(In[i].type == 4)
			{
				double d, sidemin=10;
				double temp1[3], temp2[3], temp3[3];
				double center[3];
				int edgeflags=0;
				bool nocond=true;

				d=DIST(In[i].b, In[i].a);
				if(d < sidemin) sidemin=d;
				d=DIST(In[i].c, In[i].b);
				if(d < sidemin) sidemin=d;
				d=DIST(In[i].d, In[i].c);
				if(d < sidemin) sidemin=d;
				d=DIST(In[i].a, In[i].d);
				if(d < sidemin) sidemin=d;
				if(d<20*SMALL) continue; // Too small - don't bother
				ADD(temp1, In[i].c, In[i].a);
				MULT (temp1, temp1, 0.5);
				ADD(temp2, In[i].d, In[i].b);
				MULT (temp2, temp2, 0.5);
				if(DIST(temp1, temp2) > SMALL) continue; // not a parallelogram
				// center coordinates
				ADD (center, temp1, temp2);
				MULT (center, center, 0.5);
				// Find adjacent condlines
				if(noadjacent)
				{
					for(j=0; j<NumInMax; j++)
					{
						if(In[j].type != 5) continue;
						if(DIST(In[i].a, In[j].a) < SMALL)
						{
							if((DIST(In[i].b, In[j].b) < SMALL) || (DIST(In[i].d, In[j].b) < SMALL))
							{
								nocond=false;
								continue;
							}
						}
						if(DIST(In[i].a, In[j].b) < SMALL)
						{
							if((DIST(In[i].b, In[j].a) < SMALL) || (DIST(In[i].d, In[j].a) < SMALL))
							{
								nocond=false;
								continue;
							}
						}
						if(DIST(In[i].c, In[j].a) < SMALL)
						{
							if((DIST(In[i].b, In[j].b) < SMALL) || (DIST(In[i].d, In[j].b) < SMALL))
							{
								nocond=false;
								continue;
							}
						}
						if(DIST(In[i].c, In[j].b) < SMALL)
						{
							if((DIST(In[i].b, In[j].a) < SMALL) || (DIST(In[i].d, In[j].a) < SMALL))
							{
								nocond=false;
								continue;
							}
						}
					}
				}
				
				// Find edges
				if(nocond)
				{
					for(j=0; j<NumInMax; j++)
					{
						if(In[j].type != 2) continue;
						if(DIST(In[i].a, In[j].a) < SMALL)
						{
							if(DIST(In[i].b, In[j].b) < SMALL)
							{
								In[j].type=-1;
								SubstLine++;
								edgeflags |= 1;
								continue;
							}
							if(DIST(In[i].d, In[j].b) < SMALL)
							{
								In[j].type=-1;
								SubstLine++;
								edgeflags |= 8;
								continue;
							}
						}
						if(DIST(In[i].a, In[j].b) < SMALL)
						{
							if(DIST(In[i].b, In[j].a) < SMALL)
							{
								In[j].type=-1;
								SubstLine++;
								edgeflags |= 1;
								continue;
							}
							if(DIST(In[i].d, In[j].a) < SMALL)
							{
								In[j].type=-1;
								SubstLine++;
								edgeflags |= 8;
								continue;
							}
						}
						if(DIST(In[i].c, In[j].a) < SMALL)
						{
							if(DIST(In[i].b, In[j].b) < SMALL)
							{
								In[j].type=-1;
								SubstLine++;
								edgeflags |= 2;
								continue;
							}
							if(DIST(In[i].d, In[j].b) < SMALL)
							{
								In[j].type=-1;
								SubstLine++;
								edgeflags |= 4;
								continue;
							}
						}
						if(DIST(In[i].c, In[j].b) < SMALL)
						{
							if(DIST(In[i].b, In[j].a) < SMALL)
							{
								In[j].type=-1;
								SubstLine++;
								edgeflags |= 2;
								continue;
							}
							if(DIST(In[i].d, In[j].a) < SMALL)
							{
								In[j].type=-1;
								SubstLine++;
								edgeflags |= 4;
								continue;
							}
						}
					}
				}
				In[i].rect=0;
				switch(edgeflags)
				{
				case 0:
					continue;
					break;
				case 1:
					ROTQUAD(In[i]);
					ROTQUAD(In[i]);
					ROTQUAD(In[i]);
					In[i].rect=1;
					break;
				case 2:
					In[i].rect=1;
					break;
				case 4:
					ROTQUAD(In[i]);
					In[i].rect=1;
					break;
				case 8:
					ROTQUAD(In[i]);
					ROTQUAD(In[i]);
					In[i].rect=1;
					break;
				case 3:
					ROTQUAD(In[i]);
					ROTQUAD(In[i]);
					ROTQUAD(In[i]);
					In[i].rect=2;
					break;
				case 6:
					In[i].rect=2;
					break;
				case 12:
					ROTQUAD(In[i]);
					In[i].rect=2;
					break;
				case 9:
					ROTQUAD(In[i]);
					ROTQUAD(In[i]);
					In[i].rect=2;
					break;
				case 5:
					In[i].rect=3;
					break;
				case 10:
					ROTQUAD(In[i]);
					In[i].rect=3;
					break;
				case 7:
					ROTQUAD(In[i]);
					ROTQUAD(In[i]);
					ROTQUAD(In[i]);
					In[i].rect=4;
					break;
				case 14:
					In[i].rect=4;
					break;
				case 13:
					ROTQUAD(In[i]);
					In[i].rect=4;
					break;
				case 11:
					ROTQUAD(In[i]);
					ROTQUAD(In[i]);
					In[i].rect=4;
					break;
				case 15:
					In[i].rect=5;
					break;
				}
				SubstQuad++;
				// Quad local basis
				SUB(temp1, In[i].b, In[i].a);
				SUB(temp2, In[i].d, In[i].a);
				CROSS(temp3, temp2, temp1);
// remain with "simple" values
				int index=0, k;
				double maxco=0;
				for(k=0; k<3; k++)
				{
					if(fabs(temp3[k])>maxco) 
					{
						maxco = fabs(temp3[k]);
						index=k;
					}
				}
				for(k=0; k<3; k++)
				{
					if(k==index) temp3[k]=temp3[k]/maxco; else temp3[k]=0;
				}

				if(!In[i].bfc) 
				{					
					MULT(temp3, temp3, -1.);
				}

				MULT(temp1, temp1, 0.5);
				MULT(temp2, temp2, 0.5);
				In[i].type=1;
				SET(In[i].a, center);

				In[i].b[0]=temp1[0];
				In[i].c[0]=temp1[1];
				In[i].d[0]=temp1[2];
				In[i].b[1]=temp3[0];
				In[i].c[1]=temp3[1];
				In[i].d[1]=temp3[2];
				In[i].b[2]=temp2[0];
				In[i].c[2]=temp2[1];
				In[i].d[2]=temp2[2];
				if(color) In[i].color=1;
			}
		}
	}

	if(SubstQuad > 0) printf("%d quad(s) and %d line(s) were replaced with %d rect primitive(s)\n", SubstQuad, SubstLine, SubstQuad);
	printf("Writing output file...\n");

	LdrawIn = fopen (argv[argc-2], "rt");
	if(LdrawIn == NULL)
	{
		printf("Can't open file\n%s\n", argv[argc-3]);
		printf("Press <Enter> to quit");
		getchar();
		return 2;
	}

	for(i=0; i<NumInMax; i++)
	{
		fgets(buf, 1024, LdrawIn);
		
		switch (In[i].type)
		{
		case 0:
			fputs(buf, LdrawOut);
			break;
		case 1:
			if (In[i].color == -1)
			{
				fprintf(LdrawOut, "%d 0x%x %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg", In[i].type, In[i].hex_color, In[i].a[0], In[i].a[1], In[i].a[2], In[i].b[0], In[i].b[1], In[i].b[2]);
			}
			else
			{
				fprintf(LdrawOut, "%d %d %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg", In[i].type, In[i].color, In[i].a[0], In[i].a[1], In[i].a[2], In[i].b[0], In[i].b[1], In[i].b[2]);
			}
			fprintf(LdrawOut, " %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg", In[i].c[0], In[i].c[1], In[i].c[2], In[i].d[0], In[i].d[1], In[i].d[2]);
			switch(In[i].rect)
			{
			case 1:
				fprintf(LdrawOut, " rect1.dat\n");
				break;
			case 2:
				fprintf(LdrawOut, " rect2a.dat\n");
				break;
			case 3:
				fprintf(LdrawOut, " rect2p.dat\n");
				break;
			case 4:
				fprintf(LdrawOut, " rect3.dat\n");
				break;
			case 5:
				fprintf(LdrawOut, " rect.dat\n");
				break;
			default:
				fprintf(LdrawOut, " __%d__\n",In[i].rect);
			}
			break;
		case 2:
			if (In[i].color == -1)
			{
				fprintf(LdrawOut, "%d 0x%x %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", In[i].type, In[i].hex_color, In[i].a[0], In[i].a[1], In[i].a[2], In[i].b[0], In[i].b[1], In[i].b[2]);
			}
			else
			{
				fprintf(LdrawOut, "%d %d %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", In[i].type, In[i].color, In[i].a[0], In[i].a[1], In[i].a[2], In[i].b[0], In[i].b[1], In[i].b[2]);
			}
			break;
		case 3:
			if (In[i].color == -1)
			{
				fprintf(LdrawOut, "%d 0x%x %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg", In[i].type, In[i].hex_color, In[i].a[0], In[i].a[1], In[i].a[2], In[i].b[0], In[i].b[1], In[i].b[2]);
			}
			else
			{
				fprintf(LdrawOut, "%d %d %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg", In[i].type, In[i].color, In[i].a[0], In[i].a[1], In[i].a[2], In[i].b[0], In[i].b[1], In[i].b[2]);
			}
			fprintf(LdrawOut, " %.7lg %.7lg %.7lg\n", In[i].c[0], In[i].c[1], In[i].c[2]);
			break;
		case 4:
		case 5:
			if (In[i].color == -1)
			{
				fprintf(LdrawOut, "%d 0x%x %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg", In[i].type, In[i].hex_color, In[i].a[0], In[i].a[1], In[i].a[2], In[i].b[0], In[i].b[1], In[i].b[2]);
			}
			else
			{
				fprintf(LdrawOut, "%d %d %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg", In[i].type, In[i].color, In[i].a[0], In[i].a[1], In[i].a[2], In[i].b[0], In[i].b[1], In[i].b[2]);
			}
			fprintf(LdrawOut, " %.7lg %.7lg %.7lg %.7lg %.7lg %.7lg\n", In[i].c[0], In[i].c[1], In[i].c[2], In[i].d[0], In[i].d[1], In[i].d[2]);
			break;
		}
	}

	fclose (LdrawIn);
	fclose (LdrawOut);
	printf("Press <Enter> to quit");
	getchar();
	return(0);
}


